import json
import os
import re
import shutil
from inspect import getfullargspec
from string import Template
from xml.sax import saxutils
import pandas as pd
import pkg_resources
import requests
from functools import wraps
import pysolr
from itertools import chain



# from jsonpath_rw import parse as parse_jpath
from vfb_connect.neo.neo4j_tools import chunks, Neo4jConnect, dict_cursor, escape_string

# Connect to the VFB SOLR server
vfb_solr = pysolr.Solr('http://solr.virtualflybrain.org/solr/vfb_json/', always_commit=False, timeout=990)


def batch_query(func):
    """Decorator to batch the first argument of the function (assumed to be an iterable) and apply the function to each batch.
    
    Assumes the first argument is to be batched and that the return value is a list. Only works on class methods.
    
    :param func: The function to be wrapped and batched.
    :return: A wrapper function that processes the input in batches.
    """
    @wraps(func)
    def wrapper_batch(*args, **kwargs):
        arg_names = getfullargspec(func).args

        if len(args) > 1:
            arg1v = args[1]
            arg1typ = 'a'
        else:
            arg1v = kwargs.get(arg_names[1])
            arg1typ = 'k'

        cs = chunks(arg1v, 2500)
        out = []
        for c in cs:
            if arg1typ == 'a':
                arglist = list(args)
                arglist[1] = c
                subargs = tuple(arglist)
                out.extend(func(*subargs, **kwargs))
            elif arg1typ == 'k':
                kwargdict = dict(kwargs)
                kwargdict[arg_names[1]] = c
                out.extend(func(*args, **kwargdict))
        # Check if return_dataframe is requested and summary is requested
        if out and isinstance(out, list) and len(out) > 0 and dict(kwargs).get('return_dataframe', 'return_dataframe' in arg_names) and dict(kwargs).get('summary', True):
            try:
                return pd.DataFrame.from_records(out)
            except TypeError as e:
                # Debugging information
                print("TypeError encountered during DataFrame conversion.")
                print(f"Error message: {e}")
                print("Data attempted to convert to DataFrame:")
                print(f"Data (first 5 records): {out[:5] if out else 'No data available'}")
                print("Arguments provided:")
                print(f"kwargs: {kwargs}")
                print(f"arg_names: {arg_names}")
                raise e
        return out
    return wrapper_batch



# def pop_from_jpath(jpath, json, join=True):
#     expr = parse_jpath(jpath)
#     if join:
#         return '|'.join([match.value for match in expr.find(json)])
#     else:
#         return [match.value for match in expr.find(json)]


def _populate_minimal_summary_tab(TermInfo):
    d = dict()
    d['label'] = TermInfo['term']['core']['label']
    d['symbol'] = TermInfo['term']['core'].get('symbol', '')
    d['id'] = TermInfo['term']['core']['short_form']
    d['tags'] = TermInfo['term']['core']['types']
    return d

def _populate_data_source_id(TermInfo, d = dict()):
    # Populate data source summary tab if source xrefs
    if 'xrefs' in TermInfo.keys():
        data_sources = []
        ds_accessions = []

        for p in TermInfo['xrefs']:
            # Check if 'is_data_source' is set and truthy
            if p.get('is_data_source') and p.get('is_data_source') is True:
                # Add site short_form or symbol to data_sources
                symbol = p['site'].get('symbol')
                if not symbol:  # This covers both None and empty string
                    symbol = p['site']['short_form']
                data_sources.append(symbol)
                # Add accession if available
                if 'accession' in p.keys():
                    ds_accessions.append(p['accession'])

        # Join the results with '|'
        d['data_source'] = data_sources
        d['accession'] = ds_accessions
    return d


def _populate_anatomical_entity_summary(TermInfo):
    d = _populate_minimal_summary_tab(TermInfo)
    d = _populate_data_source_id(TermInfo, d)
#   d['parents_symbol'] = pop_from_jpath("$.parents[*].symbol", TermInfo)
#   d['parents_label'] = pop_from_jpath("$.parents[*].label", TermInfo)
    d['parents_label'] = [str(p['label']) for p in TermInfo['parents']]
#   d['parents_id'] = pop_from_jpath("$.parents[*].short_form", TermInfo)
    d['parents_id'] = [str(p['short_form']) for p in TermInfo['parents']]

    return d


def _populate_instance_summary_tab(TermInfo):
    d = _populate_anatomical_entity_summary(TermInfo)
#    site_expr = "$.xrefs.[*].site.short_form"
#    acc_expr = "$.xrefs.[*].accession"
#    is_data_source_expr = "$.xrefs.[*].is_data_source"
#   sites = pop_from_jpath(site_expr, TermInfo, join=False)
    i = 0
    if 'xrefs' in TermInfo.keys():
        d['xrefs'] = [
            f"{p['site']['symbol'] if p['site'].get('symbol') else p['site']['short_form']}:{p['accession']}"
            for p in TermInfo['xrefs']
        ]
    d['templates'] = [str(x['image']['template_anatomy']['label']) for x in TermInfo['channel_image']]
    d['dataset'] = [str(x['dataset']['core']['short_form']) for x in TermInfo['dataset_license']]
    d['license'] = [str(x['license']['link']) for x in TermInfo['dataset_license']
                             if 'link' in x['license'].keys()]
    return d


def _populate_dataset_summary_tab(TermInfo):
    d = _populate_minimal_summary_tab(TermInfo)
    d['description'] = TermInfo['term']['description']
    d['miniref'] = [str(x['core']['label']) for x in TermInfo['pubs']]
    d['FlyBase'] = [str(x['FlyBase']) for x in TermInfo['pubs'] if 'FlyBase' in x]
    d['PMID'] = [str(x['PubMed']) for x in TermInfo['pubs'] if 'PubMed' in x]
    d['DOI'] = [str(x['DOI']) for x in TermInfo['pubs'] if 'PubMed' in x]
    return d


def _populate_manifest(filename, instance):
    d = _populate_instance_summary_tab(instance)
    d['filename'] = filename
    return d

def _populate_summary(TermInfo):
    """
    Generalized function to populate a summary dictionary based on the fields available in the TermInfo JSON,
    including cross-reference (xref) terms formatted as 'DB:xxxxx'.

    :param TermInfo: The JSON object containing term information.
    :rtype: dict
    """
    def get_value(key, default=''):
        return TermInfo.get(key, default)
    
    d = dict()
    
    # Populate minimal summary tab
    d['label'] = get_value('term', {}).get('core', {}).get('label', '')
    d['symbol'] = get_value('term', {}).get('core', {}).get('symbol', '')
    d['id'] = get_value('term', {}).get('core', {}).get('short_form', '')
    d['tags'] = get_value('term', {}).get('core', {}).get('types', [])

    if get_value('term', {}).get('description', '') or get_value('term', {}).get('comment', ''):
        d['description'] = ' '.join(get_value('term', {}).get('description', '')) + ' ' + ' '.join(get_value('term', {}).get('comment', ''))
    
    # Populate anatomical entity summary if available
    if 'parents' in TermInfo.keys():
        d['parents_label'] = [str(p['label']) for p in TermInfo['parents']]
        d['parents_id'] = [str(p['short_form']) for p in TermInfo['parents']]

    d = _populate_data_source_id(TermInfo, d)

    # Populate instance summary tab if available
    if 'xrefs' in TermInfo.keys():
        d['xrefs'] = [
            f"{p['site']['symbol'] if p['site'].get('symbol') else p['site']['short_form']}:{p['accession']}"
            for p in TermInfo['xrefs']
        ]

    if 'channel_image' in TermInfo.keys():
        d['templates'] = [str(x['image']['template_anatomy']['label']) for x in TermInfo['channel_image']]
    
    if 'dataset_license' in TermInfo.keys():
        d['dataset'] = [str(x['dataset']['core']['short_form']) for x in TermInfo['dataset_license']]
        d['license'] = [str(x['license']['link']) for x in TermInfo['dataset_license'] if 'link' in x['license'].keys()]

    # Populate dataset summary tab if available
    if 'term' in TermInfo.keys() & 'description' in TermInfo['term'].keys():
        d['description'] = get_value('term', {}).get('description', '')
    
    if 'pubs' in TermInfo.keys():
        d['miniref'] = [str(x['core']['label']) for x in TermInfo['pubs']]
        d['FlyBase'] = [str(x['FlyBase']) for x in TermInfo['pubs'] if 'FlyBase' in x]
        d['PMID'] = [str(x['PubMed']) for x in TermInfo['pubs'] if 'PubMed' in x]
        d['DOI'] = [str(x['DOI']) for x in TermInfo['pubs'] if 'DOI' in x]

    return d

# def filter_term_content(func):
#     """Decorator function that wraps queries that return lists of JSON objects and which have
#      an arg named filters.  The filters arg takes a filter object, which specifics JPATH queries which are applied
#      as a filter to each returned JSON object so that the final result only contains the
#      specified paths and their values"""
#
#     type_2_summary = {
#         'individual': '_populate_instance_summary_tab',
#         'class': '_populate_anatomical_entity_summary',
#     }

    # def filter_wrapper(*args, **kwargs):
    #     func_ret = func(*args, **kwargs)
    #     # arg_names = getfullargspec(func).args # Potentially useful for capturing defaults
    #                                             # but some problem getting working with
    #                                             # nested decorators
    #     out = []
    #     if ('filters' in kwargs.keys()):
    #         for f in kwargs['filters']:
    #             expr = parse_jpath(f)
    #             for fr in func_ret:
    #                 matching_fields = [match for match in expr.find(fr)]
    #                 for mf in matching_fields:
    #                     out.append(mf.value)
    #         return out
    #     else:
    #         return func_ret
    #
    # return filter_wrapper


def gen_simple_report(terms, dataframe=True):
    nc = Neo4jConnect("https://pdb.virtualflybrain.org", "neo4j", "neo4j")
    query = """MATCH (n:Class) WHERE n.iri in %s WITH n 
                OPTIONAL MATCH  (n)-[r]->(p:pub) WHERE r.typ = 'syn' 
                WITH n, 
                COLLECT({ synonym: r.synonym, PMID: 'PMID:' + p.PMID, 
                    miniref: p.label}) AS syns 
                OPTIONAL MATCH (n)-[r]-(p:pub) WHERE r.typ = 'def' 
                with n, syns, 
                collect({ PMID: 'PMID:' + p.PMID, miniref: p.label}) as pubs
                OPTIONAL MATCH (n)-[:SUBCLASSOF]->(super:Class)
                RETURN n.short_form as short_form, n.label as label, 
                n.description as description, syns, pubs,
                super.label, super.short_form
                 """ % str(terms)
    q = nc.commit_list([query])
    # add check
    if not q:
        raise Exception('Query failed.')
    if dataframe:
        return pd.DataFrame.from_records(dict_cursor(q))
    return dict_cursor(q)


class QueryWrapper(Neo4jConnect):

    def __init__(self, *args, **kwargs):
        super(QueryWrapper, self).__init__(*args, **kwargs)
        query_json = pkg_resources.resource_filename(
                            "vfb_connect",
                            "resources/VFB_TermInfo_queries.json")
        with open(query_json, 'r') as f:
            self.queries = json.loads(saxutils.unescape(f.read()))

#    def get_sites(self):
#        return

    def _query(self, q):
        qr = self.commit_list([q])
        if not qr:
            raise Exception(f'Query failed: {q}')
        else:
            r = dict_cursor(qr)
            if not r:
                print(f'\033[33mWarning:\033[0m No results returned for query: {q} -> {qr} -> {r}')
                return []
            else:
                return r

    def get_images(self, short_forms: iter, template, image_folder, image_type='swc', stomp=False):
        """Given an iterable of `short_forms` for instances, find all images of specified `image_type`
        registered to `template`. Save these to `image_folder` along with a manifest.tsv.  Return manifest as
        pandas DataFrame.

        :param short_forms: iterable (e.g. list) of VFB IDs of Individuals with images
        :param template: template name
        :param image_folder: folder to save image files & manifest to.
        :param image_type: image type (file extension)
        :param stomp: Overwrite image_folder if already exists.
        :return: Manifest as Pandas DataFrame
        """

        # TODO - make image type into array
        short_forms = list(short_forms)
        manifest = []
        if stomp and os.path.isdir(image_folder):
            if shutil.rmtree.avoids_symlink_attacks:
                shutil.rmtree(image_folder)
            else:
                print(f"\033[33mWarning:\033[0m Not deleting {image_folder}, stomp option not supported on this system for security reasons,"
                              "please delete manually.")
        os.makedirs(image_folder, exist_ok=True)
        inds = self.get_anatomical_individual_TermInfo(short_forms=short_forms, summary=False)
        if not inds:
            print(f"\033[33mWarning:\033[0m No results returned for short_forms: {short_forms}")
        else:
            print(f"Got {len(inds)} results.")
            for i in inds:
                if not ('has_image' in i['term']['core']['types']):
                    continue
                label = i['term']['core']['label']
                image_matches = [x['image'] for x in i['channel_image']]
                if not image_matches:
                    continue
                for imv in image_matches:
                    if imv['template_anatomy']['label'] == template:
                        r = requests.get(imv['image_folder'] + '/volume.' + image_type)
                        ### Slightly dodgy warning - could mask network errors
                        if not r.ok:
                            print("33mWarning:\033[0m No '%s' file found for '%s'." % (image_type, label))
                            continue
                        filename = re.sub('\W', '_', label) + '.' + image_type
                        with open(image_folder + '/' + filename, 'w') as image_file:
                            image_file.write(r.text)
                        manifest.append(_populate_manifest(instance=i, filename=filename))
        manifest_df = pd.DataFrame.from_records(manifest)
        manifest_df.to_csv(image_folder + '/manifest.tsv', sep='\t')
        return manifest_df

    def get_dbs(self, include_symbols=False):
        """Get a list of available database IDs

        :return: list of VFB IDs."""
        query = "MATCH (i:Individual) " \
                "WHERE i:Site OR i:API " \
                "return i.short_form"
        results = self._query(query)
        dbs = [d['i.short_form'] for d in results]
        if include_symbols:
            query = "MATCH (i:Individual) " \
                    "WHERE i:Site OR i:API AND exists(i.symbol) and not i.symbol[0] = '' " \
                    "RETURN i.symbol"
            results = self._query(query)
            dbs.extend([d['i.symbol'] for d in results if d['i.symbol']])
        return dbs

    def get_datasets(self, summary=True, return_dataframe=True):
        """
        Generate a report of all datasets available in the system.

        This method retrieves all datasets and returns their information as either full JSON data structures or summaries.
        The results can be returned as a list of dictionaries or as a pandas DataFrame, depending on the `return_dataframe` flag.

        :param summary: Optional. If `True`, returns summary reports instead of full metadata. Default is `True`.
        :param return_dataframe: Optional. If `True` and `summary` is also `True`, returns the results as a pandas DataFrame.
                                Default is `True`.
        :return: A list of terms as nested Python data structures (VFB_json or summary_report_json), or a pandas DataFrame if 
                `return_dataframe` is `True` and `summary` is `True`.
        :rtype: list of dicts or pandas.DataFrame
        """
        dc = self._query("MATCH (ds:DataSet) "
                        "RETURN ds.short_form AS sf")
        short_forms = [d['sf'] for d in dc]
        results = self.get_DataSet_TermInfo(short_forms, summary=summary, return_dataframe=False)
        if return_dataframe and summary:
            return pd.DataFrame.from_records(results)
        return results

    def get_templates(self, summary=True, return_dataframe=True):
        """Generate JSON report of all available templates.

            :param summary: Optional.  Returns summary reports if `True`. Default `True`
            :return: Returns a list of terms as nested python data structures following VFB_json or a summary_report_json
            :return type: list of VFB_json or summary_report_json
            """
        dc = self._query("MATCH (i:Individual:Template:Anatomy) "
                         "RETURN i.short_form as sf")
        short_forms = [d['sf'] for d in dc]
        return self.get_anatomical_individual_TermInfo(short_forms, summary=summary)

    def vfb_id_2_xrefs(self, vfb_id: iter, db='', id_type='', reverse_return=False):
        """Map a list of short_form IDs in VFB to external DB IDs

        :param vfb_id: An iterable (e.g. a list) of VFB short_form IDs.
        :param db: optional specify the VFB id (short_form) of an external DB to map to. (use get_dbs to find options)
        :param id_type: optionally specify an external id_type
        :param reverse_return: Boolean: Optional (see return)
        :return: if `reverse_return` is False:
                dict { VFB_id : [{ db: <db> : acc : <acc> }
            Return if `reverse_return` is `True`:
                dict { acc : [{ db: <db> : vfb_id : <VFB_id> }
        """
        vfb_id = list(set(vfb_id))
        match = "MATCH (s:Individual)<-[r:database_cross_reference]-(i:Entity) " \
                "WHERE i.short_form in %s" % str(vfb_id)
        clause1 = ''
        if db:
            clause1 = "AND s.short_form = '%s'" % db
        clause2 = ''
        if id_type:
            clause2 = "AND r.id_type = '%s'" % id_type
        ret = "RETURN i.short_form as key, " \
              "collect({ db: s.short_form, acc: r.accession[0]}) as mapping"
        if reverse_return:
            ret = "RETURN r.accession[0] as key, " \
                  "collect({ db: s.short_form, vfb_id: i.short_form }) as mapping"
        q = ' '.join([match, clause1, clause2, ret])
        dc = self._query(q)
        mapping = {d['key']: d['mapping'] for d in dc}
        unmapped = set(vfb_id)-set(mapping.keys())
        if unmapped:
            print("33mWarning:\033[0m The following IDs do not match DB &/or id_type constraints: %s" % str(unmapped))
        return mapping

    def vfb_id_2_neuprint_bodyID(self, vfb_id, db=''):
        mapping = self.vfb_id_2_xrefs(vfb_id, db=db, reverse_return=True)
        return [int(k) for k, v in mapping.items()]

    def xref_2_vfb_id(self, acc=None, db='', id_type='', reverse_return=False, verbose=False):
        """Map a list external DB IDs to VFB IDs

          :param acc: An iterable (e.g. a list) of external IDs (e.g. neuprint bodyIDs).
          :param db: optional specify the VFB id (short_form) of an external DB to map to. (use get_dbs to find options)
          :param id_type: optionally specify an external id_type
          :param reverse_return: Boolean: Optional (see return)
          :return: if `reverse_return` is False:
                dict { acc : [{ db: <db> : vfb_id : <VFB_id> }
              Return if `reverse_return` is `True`:
                dict { VFB_id : [{ db: <db> : acc : <acc> }
          """
        if isinstance(acc, str):
            acc = [acc]
        match = "MATCH (s:Individual)<-[r:database_cross_reference]-(i:Entity) WHERE"
        conditions = []
        if not (acc is None):
            acc = [str(a) for a in set(acc)]
            conditions.append("r.accession[0] in %s" % str(acc))
        if db:
            conditions.append("s.short_form = '%s'" % db)
        if id_type:
            conditions.append("r.id_type = '%s'" % id_type)
        condition_clauses = ' AND '.join(conditions)
        ret = "RETURN r.accession[0] as key, " \
              "collect({ db: s.short_form, vfb_id: i.short_form }) as mapping"
        if reverse_return:
            ret = "RETURN i.short_form as key, " \
                  "collect({ db: s.short_form, acc: r.accession[0]}) as mapping"
        q = ' '.join([match, condition_clauses, ret])
        print(q) if verbose else None
        dc = self._query(q)
        print(dc) if verbose else None
        return {d['key']: d['mapping'] for d in dc}

    @batch_query
    def get_terms_by_xref(self, acc, db='', id_type='', summary=True, return_dataframe=True):
        """
        Generate a JSON report for terms specified by a list of cross-reference IDs (xrefs).

        This method maps external database IDs (xrefs) to VFB (Virtual Fly Brain) IDs and then retrieves detailed term 
        information for those VFB IDs. The information can be returned either as a summary or as full term metadata.

        :param acc: An iterable (e.g., a list) of external cross-reference IDs.
        :param db: Optional. Specify the VFB ID (short_form) of an external database to map to. (Use `get_dbs()` to find options).
        :param id_type: Optional. Specify an external ID type to filter the mapping results.
        :param summary: Optional. Returns summary reports if `True`. Default is `True`.
        :param return_dataframe: Optional. Includes related datasets in the report if `True`. Default is `True`.
        :return: A list of term metadata as VFB_json or summary_report_json.
        :rtype: list of dicts
        :raises Warning: If no VFB ID is found for a given xref.
        """
        # Fetch VFB IDs associated with the xref
        vfb_ids = self.xref_2_vfb_id(acc, db=db, id_type=id_type)

        # Extract the list of IDs from the response
        ids_to_query = []
        for key in acc:
            if key not in vfb_ids.keys():
                print("33mWarning:\033[0m No VFB ID found for %s" % key)
            else:
                ids_to_query.append(vfb_ids[key][0]['vfb_id'])

        # Retrieve term information for all IDs
        return self.get_TermInfo(ids_to_query, summary=summary, return_dataframe=False)

    def get_images_by_filename(self, filenames, dataset=None, summary=True, return_dataframe=True):
        """Takes a list of filenames as input and returns a list of image terminfo.
        Optionally restrict by dataset (improves speed)"""
        m = "MATCH (ds:DataSet)<-[has_source]-(ai:Individual)<-[:depicts]" \
            "-(channel:Individual)-[irw:in_register_with]-(tc:Template)"
        w = "WHERE irw.filename[0]in %s" % str([escape_string(f) for f in filenames])
        if dataset:
            w += "AND ds.short_form = '%s'" % dataset
        r = "RETURN ai.short_form"
        dc = self._query(' '.join([m, w, r]))
        return self.get_anatomical_individual_TermInfo([d['ai.short_form']
                                                        for d in dc], summary=summary, return_dataframe=return_dataframe)

    @batch_query
    def get_TermInfo(self, short_forms: iter, summary=True, cache=True, return_dataframe=True, limit=None, verbose=False):
        """
        Generate a JSON report or summary for terms specified by a list of VFB IDs.

        This method retrieves term information for a list of specified VFB IDs (short_forms). It can return either 
        full metadata or a summary of the terms. The results can be returned as a pandas DataFrame if `return_dataframe` 
        is set to `True`.

        :param short_forms: An iterable (e.g., a list) of VFB IDs (short_forms).
        :param summary: Optional. If `True`, returns a summary report instead of full metadata. Default is `True`.
        :param cache: Optional. If `True`, attempts to retrieve cached results before querying. Default is `True`.
        :param return_dataframe: Optional. If `True`, returns the results as a pandas DataFrame. Default is `True`.
        :return: A list of term metadata as VFB_json or summary_report_json, or a pandas DataFrame if `return_dataframe` is `True`.
        :rtype: list of dicts or pandas.DataFrame
        """
        from vfb_connect import vfb
        if cache:
            result = self._get_Cached_TermInfo(short_forms, summary=summary, return_dataframe=False, verbose=verbose)
            cn = len(set(short_forms))
            rn = len(result)
            if rn != cn:
                print(f"\033[33mWarning:\033[0m Cache didn't return all results. Got {rn} out of {cn}") if verbose else None
                missing = set(short_forms) - set([r['term']['core']['short_form'] for r in result])
                print(f"Missing: {missing}") if verbose else None
                for i in missing:
                    print(f"Checking: {i}") if verbose else None
                    if not i in vfb.lookup.values():
                        print(f"\033[33mWarning:\033[0m called a non existant id:{i}")
                        cn -= 1
            if rn == cn:
                print("Using cached results.") if verbose else None
                result = result[:limit] if limit else result
                if summary:
                    results = []
                    for r in result:
                        results.append(_populate_summary(r))
                    return results
                else:
                    return result
            else:
                print(f"\033[33mWarning:\033[0m Cache didn't return all results. Got {rn} out of {cn}. Falling back to slower query.")
                return self.get_TermInfo(short_forms, summary=summary, cache=False, return_dataframe=return_dataframe, limit=limit)
        print("Pulling results from VFB PDB (Neo4j): http://pdb.virtualflybrain.org") if verbose else None
        pre_query = "MATCH (e:Entity) " \
                    "WHERE e.short_form in %s " \
                    "RETURN e.short_form as short_form, labels(e) as labs " % str(short_forms)
        r = self._query(pre_query)
        out = []
        for e in r:
            if 'class' in e['labs'] and 'Neuron' in e['labs']:
                print(f"Getting Neuron: {e['short_form']}") if verbose else None
                out.extend(self.get_neuron_class_TermInfo([e['short_form']], summary=summary, return_dataframe=False))
            elif 'class' in e['labs'] and 'Split' in e['labs']:
                print(f"Getting Split: {e['short_form']}") if verbose else None
                out.extend(self.get_split_class_TermInfo([e['short_form']], summary=summary, return_dataframe=False))
            if 'Class' in e['labs']:
                print
                out.extend(self.get_type_TermInfo([e['short_form']], summary=summary, return_dataframe=False))
            elif 'DataSet' in e['labs']:
                print(f"Getting DataSet: {e['short_form']}") if verbose else None
                out.extend(self.get_DataSet_TermInfo([e['short_form']], summary=summary, return_dataframe=False))
            elif 'License' in e['labs']:
                print(f"Getting License: {e['short_form']}") if verbose else None
                out.extend(self.get_License_TermInfo([e['short_form']], summary=summary, return_dataframe=False))
            elif 'Template' in e['labs']:
                print(f"Getting Template: {e['short_form']}") if verbose else None
                out.extend(self.get_template_TermInfo([e['short_form']], summary=summary, return_dataframe=False))
            elif 'pub' in e['labs']:
                print(f"Getting Pub: {e['short_form']}") if verbose else None
                out.extend(self.get_pub_TermInfo([e['short_form']], summary=summary, return_dataframe=False))
            elif 'Individual' in e['labs']:
                print(f"Getting Individual: {e['short_form']}") if verbose else None
                out.extend(self.get_anatomical_individual_TermInfo([e['short_form']], summary=summary, return_dataframe=False))
        print(f"Got {len(out)} results.") if verbose else None
        return out[:limit] if limit else out

    @batch_query
    def _get_Cached_TermInfo(self, short_forms: iter, summary=True, return_dataframe=True, verbose=False):
        # Flatten the list of short_forms in case it's nested
        if isinstance(short_forms, str):
            short_forms = [short_forms]
        if isinstance(short_forms, list):
            short_forms = list(chain.from_iterable(short_forms)) if any(isinstance(i, list) for i in short_forms) else short_forms
        print(f"Checking cache for results: short_forms={short_forms}") if verbose else None
        print(f"Looking for {len(short_forms)} results.") if verbose else None
        results = self._serialize_solr_output(vfb_solr.search('*', **{'fl': 'term_info','df': 'id', 'defType': 'edismax', 'q.op': 'OR','rows': len(short_forms)+10,'fq':'{!terms f=id}'+ ','.join(short_forms)}))
        print(f"Got {len(results)} results.") if verbose else None
        if len(short_forms) != len(results):
            print(f"Warning: Cache didn't return all results. Got {len(results)} out of {len(short_forms)}") if verbose else None
            missing = set(short_forms) - set([r['term']['core']['short_form'] for r in results])
            print(f"Missing: {missing}") if verbose else None
        return results



    @batch_query
    def _get_TermInfo(self, short_forms: iter, typ, show_query=False, summary=True, return_dataframe=True):
        short_forms = list(short_forms)
        sfl = "', '".join(short_forms)
        qs = Template(self.queries[typ]).substitute(ID=sfl)
        if show_query:
            print(qs)
        if summary:
            return self._termInfo_2_summary(self._query(qs), typ=typ)
        else:
            return self._query(qs)

    def _get_anatomical_individual_TermInfo_by_type(self, classification, summary=True, return_dataframe=True, limit=None, verbose=False):
        # TODO use the limit parameter
        typ = 'Get JSON for Individual:Anatomy_by_type'
        qs = Template(self.queries[typ]).substitute(ID=classification)
        if summary:
            return self._termInfo_2_summary(self._query(qs), typ='Get JSON for Individual', verbose=verbose)
        else:
            return self._query(qs)

    def _termInfo_2_summary(self, TermInfo, typ, verbose=False):
        # type_2_summary = {
        #     'Get JSON for Individual': '_populate_instance_summary_tab',
        #     'Get JSON for Class': '_populate_anatomical_entity_summary',
        # }
        dc = []
        for r in TermInfo:
            if 'Class' in typ:
                print(f"Getting Class: {r['short_form']}") if verbose else None
                dc.append(_populate_anatomical_entity_summary(r))
            elif typ == 'Get JSON for DataSet':
                print(f"Getting DataSet: {r['short_form']}") if verbose else None
                dc.append(_populate_dataset_summary_tab(r))
            else:
                print(f"Getting Individual: {r['short_form']}") if verbose else None
                dc.append(_populate_instance_summary_tab(r))
        print(f"Got {len(dc)} results.") if verbose else None
        return dc
    
    def _query_2_summary(self, TermInfo, typ, verbose=False):
        # type_2_summary = {
        #     'Get JSON for Individual': '_populate_instance_summary_tab',
        #     'Get JSON for Class': '_populate_anatomical_entity_summary',
        # }
        dc = []
        for r in TermInfo:
            if 'Class' in typ:
                print(f"Getting Class: {r['short_form']}") if verbose else None
                dc.append(_populate_anatomical_entity_summary(r))
            elif typ == 'Get JSON for DataSet':
                print(f"Getting DataSet: {r['short_form']}") if verbose else None
                dc.append(_populate_dataset_summary_tab(r))
            else:
                print(f"Getting Individual: {r['short_form']}") if verbose else None
                dc.append(_populate_instance_summary_tab(r))
        print(f"Got {len(dc)} results.") if verbose else None
        return dc

    def _serialize_solr_output(self, results):
        """
        Serialize the sanitized dictionary to JSON for all documents returned by Solr.

        :param results: The results object containing multiple documents from Solr.
        :return: A list of deserialized JSON objects.
        """
        serialized_results = []

        for doc in results.docs:
            # Ensure 'term_info' exists and is not empty
            if 'term_info' in doc and doc['term_info']:
                json_string = doc['term_info'][0]
                result = json.loads(json_string)
                serialized_results.append(result)
        return serialized_results

    def get_anatomical_individual_TermInfo(self, short_forms, summary=True, return_dataframe=True):
        """
        Generate JSON reports for anatomical individuals from a list of VFB IDs (short_forms)

        :param short_forms: An iterable (e.g. a list) of VFB IDs (short_forms) of anatomical individuals
        :param summary: Optional.  Returns summary reports if `True`. Default `True`
        :rtype: list of VFB_json or summary_report_json
        """
        return self._get_TermInfo(short_forms, typ='Get JSON for Individual', summary=summary, return_dataframe=return_dataframe)

    def get_type_TermInfo(self, short_forms, summary=True, return_dataframe=True):
        """
        Generate JSON reports for types from a list of VFB IDs (short_forms) of classes/types.

        :param short_forms: An iterable (e.g. a list) of VFB IDs (short_forms) of types
        :param summary: Optional.  Returns summary reports if `True`. Default `True`
        :rtype: list of VFB_json or summary_report_json
        """
        return self._get_TermInfo(short_forms, typ='Get JSON for Class', summary=summary, return_dataframe=return_dataframe)

    def get_neuron_class_TermInfo(self, short_forms, summary=True, return_dataframe=True):
        """
        Generate JSON reports for neuron classes from a list of VFB IDs (short_forms) of neuron classes.

        :param short_forms: An iterable (e.g. a list) of VFB IDs (short_forms) of neuron classes
        :param summary: Optional.  Returns summary reports if `True`. Default `True`
        :rtype: list of VFB_json or summary_report_json
        """
        return self._get_TermInfo(short_forms, typ='Get JSON for Neuron Class', summary=summary, return_dataframe=return_dataframe)

    def get_split_class_TermInfo(self, short_forms, summary=True, return_dataframe=True):
        """
        Generate JSON reports for split classes from a list of VFB IDs (short_forms) of split classes.

        :param short_forms: An iterable (e.g. a list) of VFB IDs (short_forms) of split classes
        :param summary: Optional.  Returns summary reports if `True`. Default `True`
        :rtype: list of VFB_json or summary_report_json
        """
        return self._get_TermInfo(short_forms, typ='Get JSON for Split Class', summary=summary, return_dataframe=return_dataframe)

    def get_DataSet_TermInfo(self, short_forms, summary=True, return_dataframe=True):
        """
        Generate JSON reports for datasets from a list of VFB IDs (short_forms) of datasets.

        :param short_forms: An iterable (e.g. a list) of VFB IDs (short_forms) of datasets
        :param summary: Optional.  Returns summary reports if `True`. Default `True`
        :rtype: list of VFB_json or summary_report_json
        """
        return self._get_TermInfo(short_forms, typ='Get JSON for DataSet', summary=summary, return_dataframe=return_dataframe)

    def get_license_TermInfo(self, short_forms, summary=True, return_dataframe=True):
        """
        Generate JSON reports for licenses from a list of VFB IDs (short_forms) of licenses.

        :param short_forms: An iterable (e.g. a list) of VFB IDs (short_forms) of licenses
        :param summary: Optional.  Returns summary reports if `True`. Default `True`
        :rtype: list of VFB_json or summary_report_json
        """
        return self._get_TermInfo(short_forms, typ='Get JSON for License', summary=summary, return_dataframe=return_dataframe)

    def get_template_TermInfo(self, short_forms, summary=True, return_dataframe=True):
        """
        Generate JSON reports for templates from a list of VFB IDs (short_forms) of templates.

        :param short_forms: An iterable (e.g. a list) of VFB IDs (short_forms) of templates
        :param summary: Optional.  Returns summary reports if `True`. Default `True`
        :rtype: list of VFB_json or summary_report_json
        """
        return self._get_TermInfo(short_forms, typ='Get JSON for Template', summary=summary, return_dataframe=return_dataframe)

    def get_pub_TermInfo(self, short_forms, summary=True, return_dataframe=True):
        """
        Generate JSON reports for publications from a list of VFB IDs (short_forms) of publications.

        :param short_forms: An iterable (e.g. a list) of VFB IDs (short_forms) of publications
        :param summary: Optional.  Returns summary reports if `True`. Default `True`
        :rtype: list of VFB_json or summary_report_json
        """
        return self._get_TermInfo(short_forms, typ='Get JSON for pub', summary=summary, return_dataframe=return_dataframe)

    def get_anatomy_by_type_TermInfo(self, short_forms, summary=True, return_dataframe=True):
        """
        Generate JSON reports for anatomical individuals by type from a list of VFB IDs (short_forms).

        :param short_forms: An iterable (e.g. a list) of VFB IDs (short_forms) of anatomical types
        :param summary: Optional.  Returns summary reports if `True`. Default `True`
        :rtype: list of VFB_json or summary_report_json
        """
        return self._get_TermInfo(short_forms, typ='Get JSON for Individual:Anatomy_by_type', summary=summary, return_dataframe=return_dataframe)




import unittest
import time
from vfb_connect.schema.vfb_term import create_vfbterm_from_json, VFBTerms, VFBTerm, Score, Relations, Xref, ExpressionList, Expression

class VfbTermTest(unittest.TestCase):

    def setUp(self):
        from vfb_connect import vfb
        self.vfb = vfb
        self.vfb._load_limit = 10

    def test_create_vfbterm_from_json(self):
        self.assertTrue(
            create_vfbterm_from_json(self.vfb.get_TermInfo("VFB_jrcv0jvf", summary=False)))

    def test_load_skeleton(self):
        term = self.vfb.term("VFB_jrcv0jvf")
        print("got VFBTerm ", term)
        term.load_skeleton()
        print("got skeleton ", term._skeleton)
        self.assertTrue(term._skeleton and term._skeleton.id == "VFB_jrcv0jvf")

    def test_load_mesh(self):
        term = self.vfb.term("VFB_jrcv0jvf")
        print("got VFBTerm ", term[0])
        term[0].load_mesh(verbose=True)
        print("got mesh ", term[0]._mesh)
        self.assertTrue(term[0]._mesh and term[0]._mesh.id == "VFB_jrcv0jvf")

    def test_load_volume(self):
        term = self.vfb.term("VFB_jrcv0jvf")
        print("got VFBTerm ", term[0])
        print("nrrd ", term[0].channel_images[0].image.image_nrrd)
        term[0].load_volume(verbose=True)
        print("got volume ", term[0]._volume)
        self.assertTrue(term[0]._volume and term[0]._volume.id == "VFB_jrcv0jvf")

    def test_VFBterms_by_region(self):
        oid = self.vfb.lookup_id('overlaps', return_curie=True)
        print(f"got overlaps: {oid}")
        self.assertEqual(oid, 'RO:0002131')
        terms = create_vfbterm_from_json(self.vfb.get_terms_by_region("nodulus", summary=False, verbose=True), verbose=True)
        print(f"got {len(terms)} VFBTerms: {terms}")
        self.assertTrue(terms)
        ids = terms.get_ids()
        print(f"got {len(ids)} ids: {ids}")
        self.assertTrue(len(ids) == len(terms))
        names = terms.get_names()
        print(f"got {len(names)} names: {names}")
        self.assertTrue(len(names) == len(terms))

    def test_VFBterms_plot3d(self):
        terms = self.vfb.terms(['VFB_jrchjwj7', 'VFB_jrchjwim', 'VFB_00004023', 'VFB_jrchk3b0', 'VFB_jrchk3b1', 'VFB_jrchk3b2', 'VFB_jrchk3b3', 'VFB_jrchk3b4', 'VFB_jrchk3b5', 'VFB_jrchk3b6', 'VFB_jrchk3b7', 'VFB_jrchk3b8', 'VFB_jrchk3b9', 'VFB_00007403', 'VFB_jrchk3ao', 'VFB_jrchk3ap', 'VFB_jrchk3aq', 'VFB_jrchk3ar', 'VFB_jrchk3as', 'VFB_jrchk3at', 'VFB_jrchk3au', 'VFB_jrchk3av', 'VFB_jrchk3aw', 'VFB_jrchk3ax', 'VFB_jrchk3ay', 'VFB_jrchk3az', 'VFB_jrchk3ba', 'VFB_jrchk3bb', 'VFB_jrchk3bc', 'VFB_jrchk3bd', 'VFB_jrchk3be', 'VFB_jrchk3bf', 'VFB_jrchk3bg', 'VFB_jrchk3bh', 'VFB_jrchk3bi', 'VFB_jrchk3bj', 'VFB_jrchk3bk', 'VFB_jrchk3bl', 'VFB_jrchk3bm', 'VFB_jrchk3bn', 'VFB_jrchk3bo', 'VFB_jrchk3bp', 'VFB_jrchk3bq', 'VFB_jrchk3br', 'VFB_jrchk3bs', 'VFB_jrchk3bt', 'VFB_jrchk3e0', 'VFB_jrchk3e1', 'VFB_jrchk3e2', 'VFB_jrchk3e3', 'VFB_jrchk3e4', 'VFB_jrchk3e5', 'VFB_001012bm', 'VFB_001012bk', 'VFB_001012bj', 'VFB_001012bi', 'VFB_001012bh', 'VFB_00013165', 'VFB_001001dr', 'VFB_00005531', 'VFB_00007701', 'VFB_jrchk8iq', 'VFB_jrchk3gx', 'VFB_jrchk3gy', 'VFB_jrchk3gz', 'VFB_jrchk3ha', 'VFB_jrchk3hb', 'VFB_jrchk3hc', 'VFB_jrchk3hd', 'VFB_jrchk3he', 'VFB_jrchk3hf', 'VFB_jrchk3hg', 'VFB_jrchk3hh', 'VFB_jrchk3hi', 'VFB_jrchk3hj', 'VFB_jrchk3hk', 'VFB_jrchk3hl', 'VFB_jrchk3hm', 'VFB_jrchk3hn', 'VFB_jrchk3j0', 'VFB_jrchk3ho', 'VFB_jrchk3j1', 'VFB_00005875', 'VFB_jrchk3j2', 'VFB_jrchk3hp', 'VFB_jrchk3hq', 'VFB_jrchk3j3', 'VFB_jrchk3hr', 'VFB_jrchk3j4', 'VFB_jrchk3j5', 'VFB_jrchk3hs', 'VFB_jrchk3ht', 'VFB_jrchk3j6', 'VFB_jrchk3j7', 'VFB_jrchk3hu', 'VFB_jrchk3j8', 'VFB_jrchk3hv', 'VFB_jrchk3j9', 'VFB_jrchk3hw', 'VFB_jrchk3hx'])
        terms = (terms[1:10:]+terms[0:10:])[0:2]
        terms = (terms[1:10:]+terms[0:10:])[0:2]
        self.assertTrue(isinstance(terms, VFBTerms))
        self.assertTrue(len(terms) > 0)
        self.assertTrue(isinstance(terms, VFBTerms))
        try:
            terms.plot3d(template='JRC2018Unisex', verbose=True, limit = 2)
        except Exception as e:
            print("plot3d expectedly failed with ", e)
        self.assertTrue([True for term in terms if hasattr(term, 'skeleton') or hasattr(term, 'mesh') or hasattr(term, 'volume')])

    def test_VFBterms_plot2d(self):
        terms = self.vfb.terms(['VFB_00000001','VFB_00010001'])
        self.assertTrue(isinstance(terms, VFBTerms))
        self.assertTrue(len(terms) > 0)
        self.assertTrue(isinstance(terms, VFBTerms))
        try:
            terms.plot2d(template='JRC2018Unisex', verbose=True)
        except Exception as e:
            print("plot2d expectedly failed with ", e)
        self.assertTrue([True for term in terms if hasattr(term, 'skeleton') or hasattr(term, 'mesh') or hasattr(term, 'volume')])

    def test_VFBterms_addition(self):
        terms = self.vfb.terms(['VFB_jrchjwj7', 'VFB_jrchjwim', 'VFB_00004023', 'VFB_jrchk3b0', 'VFB_jrchk3b1', 'VFB_jrchk3b2', 'VFB_jrchk3b3', 'VFB_jrchk3b4', 'VFB_jrchk3b5', 'VFB_jrchk3b6', 'VFB_jrchk3b7', 'VFB_jrchk3b8', 'VFB_jrchk3b9', 'VFB_00007403', 'VFB_jrchk3ao', 'VFB_jrchk3ap', 'VFB_jrchk3aq', 'VFB_jrchk3ar', 'VFB_jrchk3as', 'VFB_jrchk3at', 'VFB_jrchk3au', 'VFB_jrchk3av', 'VFB_jrchk3aw', 'VFB_jrchk3ax', 'VFB_jrchk3ay', 'VFB_jrchk3az', 'VFB_jrchk3ba', 'VFB_jrchk3bb', 'VFB_jrchk3bc', 'VFB_jrchk3bd', 'VFB_jrchk3be', 'VFB_jrchk3bf', 'VFB_jrchk3bg', 'VFB_jrchk3bh', 'VFB_jrchk3bi', 'VFB_jrchk3bj', 'VFB_jrchk3bk', 'VFB_jrchk3bl', 'VFB_jrchk3bm', 'VFB_jrchk3bn', 'VFB_jrchk3bo', 'VFB_jrchk3bp', 'VFB_jrchk3bq', 'VFB_jrchk3br', 'VFB_jrchk3bs', 'VFB_jrchk3bt', 'VFB_jrchk3e0', 'VFB_jrchk3e1', 'VFB_jrchk3e2', 'VFB_jrchk3e3', 'VFB_jrchk3e4', 'VFB_jrchk3e5', 'VFB_001012bm', 'VFB_001012bk', 'VFB_001012bj', 'VFB_001012bi', 'VFB_001012bh', 'VFB_00013165', 'VFB_001001dr', 'VFB_00005531', 'VFB_00007701', 'VFB_jrchk8iq', 'VFB_jrchk3gx', 'VFB_jrchk3gy', 'VFB_jrchk3gz', 'VFB_jrchk3ha', 'VFB_jrchk3hb', 'VFB_jrchk3hc', 'VFB_jrchk3hd', 'VFB_jrchk3he', 'VFB_jrchk3hf', 'VFB_jrchk3hg', 'VFB_jrchk3hh', 'VFB_jrchk3hi', 'VFB_jrchk3hj', 'VFB_jrchk3hk', 'VFB_jrchk3hl', 'VFB_jrchk3hm', 'VFB_jrchk3hn', 'VFB_jrchk3j0', 'VFB_jrchk3ho', 'VFB_jrchk3j1', 'VFB_00005875', 'VFB_jrchk3j2', 'VFB_jrchk3hp', 'VFB_jrchk3hq', 'VFB_jrchk3j3', 'VFB_jrchk3hr', 'VFB_jrchk3j4', 'VFB_jrchk3j5', 'VFB_jrchk3hs', 'VFB_jrchk3ht', 'VFB_jrchk3j6', 'VFB_jrchk3j7', 'VFB_jrchk3hu', 'VFB_jrchk3j8', 'VFB_jrchk3hv', 'VFB_jrchk3j9', 'VFB_jrchk3hw', 'VFB_jrchk3hx'], verbose=True)
        # test addition of slices
        terms = terms[1:10:]+terms[0:10:]
        self.assertTrue(isinstance(terms, VFBTerms))
        self.assertTrue(len(terms) > 0)
        # check simple slicing
        self.assertTrue(isinstance(terms[0:2], VFBTerms))
        # test addition uniqueness contraint
        terms=terms[0:2]+terms[0:2]
        self.assertTrue(len(terms)==len(terms[0:2]))

    def test_VFBterms_subtraction(self):
        terms = self.vfb.terms(['VFB_jrchjwj7', 'VFB_jrchjwim', 'VFB_00004023', 'VFB_jrchk3b0', 'VFB_jrchk3b1', 'VFB_jrchk3b2', 'VFB_jrchk3b3', 'VFB_jrchk3b4', 'VFB_jrchk3b5', 'VFB_jrchk3b6', 'VFB_jrchk3b7', 'VFB_jrchk3b8', 'VFB_jrchk3b9', 'VFB_00007403', 'VFB_jrchk3ao', 'VFB_jrchk3ap', 'VFB_jrchk3aq', 'VFB_jrchk3ar', 'VFB_jrchk3as', 'VFB_jrchk3at', 'VFB_jrchk3au', 'VFB_jrchk3av', 'VFB_jrchk3aw', 'VFB_jrchk3ax', 'VFB_jrchk3ay', 'VFB_jrchk3az', 'VFB_jrchk3ba', 'VFB_jrchk3bb', 'VFB_jrchk3bc', 'VFB_jrchk3bd', 'VFB_jrchk3be', 'VFB_jrchk3bf', 'VFB_jrchk3bg', 'VFB_jrchk3bh', 'VFB_jrchk3bi', 'VFB_jrchk3bj', 'VFB_jrchk3bk', 'VFB_jrchk3bl', 'VFB_jrchk3bm', 'VFB_jrchk3bn', 'VFB_jrchk3bo', 'VFB_jrchk3bp', 'VFB_jrchk3bq', 'VFB_jrchk3br', 'VFB_jrchk3bs', 'VFB_jrchk3bt', 'VFB_jrchk3e0', 'VFB_jrchk3e1', 'VFB_jrchk3e2', 'VFB_jrchk3e3', 'VFB_jrchk3e4', 'VFB_jrchk3e5', 'VFB_001012bm', 'VFB_001012bk', 'VFB_001012bj', 'VFB_001012bi', 'VFB_001012bh', 'VFB_00013165', 'VFB_001001dr', 'VFB_00005531', 'VFB_00007701', 'VFB_jrchk8iq', 'VFB_jrchk3gx', 'VFB_jrchk3gy', 'VFB_jrchk3gz', 'VFB_jrchk3ha', 'VFB_jrchk3hb', 'VFB_jrchk3hc', 'VFB_jrchk3hd', 'VFB_jrchk3he', 'VFB_jrchk3hf', 'VFB_jrchk3hg', 'VFB_jrchk3hh', 'VFB_jrchk3hi', 'VFB_jrchk3hj', 'VFB_jrchk3hk', 'VFB_jrchk3hl', 'VFB_jrchk3hm', 'VFB_jrchk3hn', 'VFB_jrchk3j0', 'VFB_jrchk3ho', 'VFB_jrchk3j1', 'VFB_00005875', 'VFB_jrchk3j2', 'VFB_jrchk3hp', 'VFB_jrchk3hq', 'VFB_jrchk3j3', 'VFB_jrchk3hr', 'VFB_jrchk3j4', 'VFB_jrchk3j5', 'VFB_jrchk3hs', 'VFB_jrchk3ht', 'VFB_jrchk3j6', 'VFB_jrchk3j7', 'VFB_jrchk3hu', 'VFB_jrchk3j8', 'VFB_jrchk3hv', 'VFB_jrchk3j9', 'VFB_jrchk3hw', 'VFB_jrchk3hx'])
        print('starting subtraction test with ', len(terms), ' terms')
        self.assertTrue(isinstance(terms, VFBTerms))
        self.assertTrue(len(terms) > 0)
        # test subtraction
        minus_terms=terms[0:5]-terms[0:2]
        print(f"looking for {terms[0:5]}")
        print(f"minus {terms[0:2]}")
        print(f"equals {terms[2:5]}")
        self.assertTrue(isinstance(terms, VFBTerms))
        self.assertTrue(len(terms) > 0)
        print(f"looking for {len(terms[0:5])} - {len(terms[0:2])} = {len(terms[2:5])} terms")
        print(f"got {len(minus_terms)} terms: {minus_terms}")
        self.assertTrue(len(minus_terms)==len(terms[2:5]))

    def test_create_vfbterm_from_id(self):
        term=VFBTerm("VFB_jrcv0jvf")
        print("got term ", term)
        self.assertTrue(term)
        self.assertTrue(isinstance(term, VFBTerm))

    def test_create_vfbterm_from_name(self):
        term=VFBTerm("nodulus")
        print("got term ", term)
        self.assertTrue(term)
        self.assertTrue(isinstance(term, VFBTerm))

    def test_create_vfbterms_from_list(self):
        terms=VFBTerms(["nodulus", "VFB_jrcv0jvf"])
        print("got terms ", terms)
        self.assertTrue(terms)
        self.assertTrue(isinstance(terms, VFBTerms))
        self.assertTrue(len(terms)==2)

    def test_vfbterm_subparts(self):
        term = self.vfb.term('medulla')
        print("got term ", term)
        subparts = term.subparts
        print("got subparts ", subparts)
        self.assertTrue(subparts)
        self.assertTrue(isinstance(subparts, VFBTerms))
        self.assertEqual(len(subparts), self.vfb._load_limit)

    def test_vfbterm_subtypes(self):
        term = self.vfb.term('FBt')
        print("got term ", term)
        subtypes = term.subtypes
        print("got subtypes ", subtypes)
        self.assertTrue(subtypes)
        self.assertTrue(isinstance(subtypes, VFBTerms))
        self.assertEqual(len(subtypes), self.vfb._load_limit)

    def test_vfbterm_children(self):
        term = self.vfb.term('medulla')
        print("got term ", term)
        children = term.children
        print("got children ", children)
        self.assertTrue(children)
        self.assertTrue(isinstance(children, VFBTerms))
        self.assertEqual(len(children), self.vfb._load_limit)

    def test_vfbterm_similarity_neuron_nblast(self):
        term = self.vfb.term('VGlut-F-000118')
        print("got term types", term.term.core.types)
        similar = term.similar_neurons_nblast
        print("got similar ", similar)
        self.assertTrue(similar)
        self.assertTrue(isinstance(similar[0], Score))
        self.assertGreater(len(similar), 50)

    # def test_vfbterm_similarity_neuron_neuronbridge(self):
    #     # TODO No neuronbridge neuron - neuron score exist at the monent
    #     term = self.vfb.term('LPC1 (FlyEM-HB:1838269993)')
    #     print("got term ", term)
    #     similar = term.similar_neurons_neuronbridge
    #     print("got similar ", similar)
    #     self.assertFalse(similar)
    #     # self.assertFalse(isinstance(similar[0], Score))
    #     self.assertFalse(len(similar)>2)

    def test_vfbterm_potental_drivers_nblast(self):
        term = self.vfb.term('VGlut-F-000118')
        print("got term ", term)
        drivers = term.potential_drivers_nblast
        print("got drivers ", drivers)
        print(f"got {len(drivers)} drivers")
        self.assertTrue(drivers)
        self.assertTrue(isinstance(drivers, list))
        self.assertTrue(isinstance(drivers[0], Score))
        self.assertTrue(len(drivers)>2)

    def test_vfbterm_potential_drivers_neuronbridge(self):
        term = self.vfb.term('LPC1 (FlyEM-HB:1838269993)')
        print("got term ", term)
        drivers = term.potential_drivers_neuronbridge
        print("got drivers ", drivers)
        print(f"got {len(drivers)} drivers")
        self.assertTrue(drivers)
        self.assertTrue(isinstance(drivers, list))
        self.assertTrue(isinstance(drivers[0], Score))
        self.assertTrue(len(drivers)>2)

    def test_vfbterm_downstream_types(self):
        term = self.vfb.term('medulla')
        print("got term ", term)
        downstream = term.downstream_neuron_types
        print("got downstream ", downstream)
        self.assertTrue(downstream)
        self.assertTrue(isinstance(downstream, VFBTerms))
        self.assertEqual(len(downstream), self.vfb._load_limit)

    def test_vfbterm_upstream_types(self):
        term = self.vfb.term('medulla')
        print("got term ", term)
        upstream = term.upstream_neuron_types
        print("got upstream ", upstream)
        self.assertTrue(upstream)
        self.assertTrue(isinstance(upstream, VFBTerms))
        self.assertEqual(len(upstream), self.vfb._load_limit)

    def test_vfbterm_overlaps_types(self):
        term = self.vfb.term('medulla')
        print("got term ", term)
        overlaps = term.neuron_types_that_overlap
        print("got overlaps ", overlaps)
        self.assertTrue(overlaps)
        self.assertTrue(isinstance(overlaps, VFBTerms))
        self.assertEqual(len(overlaps), self.vfb._load_limit)

    def test_vfbterm_overlaps_instances(self):
        term = self.vfb.term('medulla')
        print("got term ", term)
        overlaps = term.neurons_that_overlap
        print("got overlaps ", overlaps)
        self.assertTrue(overlaps)
        self.assertTrue(isinstance(overlaps, VFBTerms))
        self.assertEqual(len(overlaps), self.vfb._load_limit)

    def test_vfbterm_downstream_neurons(self):
        term = self.vfb.term('medulla')
        print("got term ", term)
        downstream = term.downstream_neurons
        print("got downstream ", downstream)
        self.assertTrue(downstream)
        self.assertTrue(isinstance(downstream, VFBTerms))
        self.assertEqual(len(downstream), self.vfb._load_limit)

    def test_vfbterm_upstream_neurons(self):
        term = self.vfb.term('medulla')
        print("got term ", term)
        upstream = term.upstream_neurons
        print("got upstream ", upstream)
        self.assertTrue(upstream)
        self.assertTrue(isinstance(upstream, VFBTerms))
        self.assertEqual(len(upstream), self.vfb._load_limit)

    def test_vfbterm_downstream_neuron_types(self):
        term = self.vfb.term('medulla')
        print("got term ", term)
        downstream_types = term.downstream_neuron_types
        print("got downstream types ", downstream_types)
        self.assertTrue(downstream_types)
        self.assertTrue(isinstance(downstream_types, VFBTerms))
        self.assertEqual(len(downstream_types), self.vfb._load_limit)

    def test_vfbterm_neuron_types_with_synaptic_terminals_here(self):
        term = self.vfb.term('medulla')
        print("got term ", term)
        synaptic_terminals_types = term.neuron_types_with_synaptic_terminals_here
        print("got synaptic terminals types ", synaptic_terminals_types)
        self.assertTrue(synaptic_terminals_types)
        self.assertTrue(isinstance(synaptic_terminals_types, VFBTerms))
        self.assertEqual(len(synaptic_terminals_types), self.vfb._load_limit)

    def test_vfbterm_neurons_with_synaptic_terminals_here(self):
        term = self.vfb.term('medulla')
        print("got term ", term)
        synaptic_terminals_neurons = term.neurons_with_synaptic_terminals_here
        print("got synaptic terminals neurons ", synaptic_terminals_neurons)
        self.assertTrue(synaptic_terminals_neurons)
        self.assertTrue(isinstance(synaptic_terminals_neurons, VFBTerms))
        self.assertEqual(len(synaptic_terminals_neurons), self.vfb._load_limit)

    def test_vfbterm_downstream_neuron_types_from_ind(self):
        term = self.vfb.term('ME on JRC2018Unisex adult brain')
        print("got term ", term)
        downstream_types = term.downstream_neuron_types
        print("got downstream types ", downstream_types)
        self.assertTrue(downstream_types)
        self.assertTrue(isinstance(downstream_types, VFBTerms))
        self.assertEqual(len(downstream_types), self.vfb._load_limit)

    def test_vfbterm_scRNAseq_Clusters(self):
        term = self.vfb.term('scRNAseq_2018_Davie_FULL_seq_clustering_dopaminergic_PAM_neurons')
        print("got term ", term)
        print(term.get_summary(return_dataframe=False))
        print(term.related_terms)
        self.assertTrue(term)

    def test_vfbterm_related_terms(self):
        term = self.vfb.term('scRNAseq_2018_Davie_FULL_seq_clustering_dopaminergic_PAM_neurons')
        print("got term ", term)
        self.assertTrue(term)
        rel = term.related_terms
        print(rel.summary)
        self.assertTrue(rel)
        self.assertTrue(isinstance(rel, Relations))
        print(term.related_terms.get_summary(return_dataframe=False))
        print(term.related_terms.get_terms())
        print(term.related_terms.get_relations())
        composed_of = term.related_terms.where(relation='composed primarily of')
        print(composed_of.get_summaries(return_dataframe=False))
        self.assertTrue(isinstance(composed_of, VFBTerms))
        term2 = self.vfb.term('VFB_jrcv0jj4', verbose=True)
        rel = term2.related_terms
        print(rel.summary)
        print(rel.where(relation='capable of').get_summaries(return_dataframe=False))
        print(dir(term2))
        print(term2.part_of)
        self.assertGreater(len(term2.part_of), 1)

    def test_vfbterm_addition(self):
        terms1 = self.vfb.terms(['medulla on JRC2018Unisex adult brain', 'nodulus on JRC2018Unisex adult brain'])
        ids = self.vfb.get_instances("'neuron' that 'has synaptic terminal in' some 'nodulus' and 'has synaptic terminal in' some 'medulla'")
        terms2 = self.vfb.terms(ids)
        print("got terms1 ", terms1)
        print("got terms2 ", terms2)
        terms = terms1+terms2
        print("got terms ", terms)
        self.assertTrue(terms)
        self.assertTrue(isinstance(terms, VFBTerms))
        self.assertTrue(len(terms) == len(terms1)+len(terms2))
        terms = terms1+terms1
        print("got terms ", terms)
        self.assertTrue(terms)
        self.assertTrue(isinstance(terms, VFBTerms))
        self.assertTrue(len(terms) == len(terms1))

        med_neurons = terms1[0].neurons_with_synaptic_terminals_here
        nod_neurons = terms1[1].neurons_with_synaptic_terminals_here
        print("got med_neurons ", med_neurons)
        print("got nod_neurons ", nod_neurons)
        # neurons = med_neurons.AND(nod_neurons)
        # print("got neurons ", neurons)
        # print(f"got {len(neurons)} via logic and {len(terms2)} via OWL query")
        # print(f"{neurons.get_names()} == {terms2.get_names()}")
        # self.assertTrue(neurons == terms2)

    def test_vfbterms_xrefs(self):
        self.vfb._dbs = None
        terms = self.vfb.terms(['catmaid_l1em:17545695', 'neuprint_JRC_Hemibrain_1point1:2039100722'])
        print("got terms ", terms)
        print(self.vfb._dbs)
        self.assertEqual(len(terms), 2)
        xref = terms[0].xrefs
        print("got xref ", xref)
        self.assertTrue(xref)
        self.assertTrue(isinstance(xref[0], Xref))
        self.assertGreaterEqual(len(xref), 1)

    def test_vfbterms_transgene_expression(self):
        term = self.vfb.term('medulla')
        print("got terms ", term)
        lct = term.lineage_clone_types
        print(lct.summary)
        self.assertTrue(lct)
        self.assertGreater(len(lct), 4)
        lc = term.lineage_clones
        print(lc.summary)
        self.assertGreater(len(lc), 9)

    def test_vfbterms_innervating(self):
        term = self.vfb.term('medulla')
        print("got terms ", term)
        lct = term.innervating
        print(lct.summary)
        self.assertTrue(lct)
        self.assertGreater(len(lct), 2)

    def test_vfbterms_transcriptomic_profile(self):
        print(self.vfb.lookup_id('LC12'))
        term = self.vfb.term('LC12')
        print("got terms ", term)
        lct = term.get_transcriptomic_profile()
        print(type(lct))
        self.assertTrue(isinstance(lct, ExpressionList))
        print(lct.summary)
        print(len(lct[0]))
        self.assertGreater(len(lct), 2)

    def test_lookups(self):
        id = self.vfb.lookup_id('LC12')
        print("got id ", id)
        self.assertEqual(id, 'FBbt_00100484')
        id = self.vfb.lookup_id('overlaps', return_curie=True)
        print("got id ", id)
        self.assertEqual(id, 'RO:0002131')
        id = self.vfb.lookup_id('cell')
        print("got id ", id)
        self.assertEqual(id, 'FBbt_00007002')
        id = self.vfb.lookup_id('part_of')
        print("got id ", id)
        self.assertEqual(id, 'BFO_0000050')

    def test_lookup_names(self):
        name = self.vfb.lookup_name('FBbt_00100484')
        print("got name ", name)
        self.assertEqual(name, 'Lcn12')
        name = self.vfb.lookup_name(["RO_0002292", "RO_0002120"])
        print("got name ", name)
        self.assertEqual(name, ['expresses', 'synapsed to'])

    def test_lookups_matching(self):
        id = self.vfb.lookup_id(' LC12' ,verbose=True)
        print("got id ", id)
        self.assertEqual(id, 'FBbt_00100484')
        id = self.vfb.lookup_id('LC_12')
        print("got id ", id)
        self.assertEqual(id, 'FBbt_00100484')
        id = self.vfb.lookup_id('LC_12 ')
        print("got id ", id)
        self.assertEqual(id, 'FBbt_00100484')
        id = self.vfb.lookup_id('LC 12 ')
        print("got id ", id)
        self.assertEqual(id, 'FBbt_00100484')

    def test_vfbterm_cache(self):
        base = len(self.vfb._term_cache)
        print(self.vfb._term_cache)

        self.vfb._use_cache=True

        # Timing the first call to 'medulla'
        start_time = time.time()
        self.vfb.term('medulla')
        end_time = time.time()
        print(f"Time taken for vfb.term('medulla'): {end_time - start_time:.4f} seconds")
        base = len(self.vfb._term_cache)
        print(base)

        # Timing the second call to 'medulla'
        start_time = time.time()
        self.vfb.term('medulla')
        end_time = time.time()
        print(f"Time taken for vfb.term('medulla') again: {end_time - start_time:.4f} seconds")

        print(base)
        self.assertEqual(len(self.vfb._term_cache), base)

        # Timing the call to 'LC12'
        start_time = time.time()
        self.vfb.term('LC12')
        end_time = time.time()
        print(f"Time taken for vfb.term('LC12'): {end_time - start_time:.4f} seconds")
        print(self.vfb._term_cache)
        self.assertGreater(len(self.vfb._term_cache), base)
        base = len(self.vfb._term_cache)
        print(base)

        self.vfb._use_cache=False

    def test_vfbterms_get_all(self):
        terms = self.vfb.terms(['IN13A015_T1_R (MANC:81202)','SNta24_MesoLN_R (MANC:45077)'])
        print("got terms ", terms)
        self.assertTrue(terms)
        self.assertTrue(isinstance(terms, VFBTerms))
        self.assertEqual(len(terms), 2)
        NT = terms.get_all('capable_of')
        print(NT)
        self.assertEqual(len(NT), 2)
        fail = terms.get_all('magic_term')
        print(fail)
        self.assertFalse(fail)
        self.assertEqual(len(fail), 0)
        NT = terms.get_all('capable_of', return_dict=True)
        print(NT)
        self.assertEqual(NT['glutamate secretion, neurotransmission'], ['VFB_jrcv1qnm'])

    def test_vfbterms_get_colours_for_terms(self):
        terms = self.vfb.terms(['IN13A015_T1_R (MANC:81202)','SNta24_MesoLN_R (MANC:45077)','SNta24_MesoLN_L (MANC:157685)','SNta24_MesoLN_L (MANC:22390588623)'])
        print("got terms ", terms)
        cp = terms.get_colours_for('capable_of')
        print(cp)
        self.assertTrue(cp)
        self.assertEqual(len(cp), 4)
        tp = terms.get_colours_for('types')
        print(tp)
        self.assertEqual(len(tp), 4)
        tp = terms.get_colours_for('types', take_first=True)
        print(tp)
        self.assertEqual(len(tp), 4)
        tp = terms.get_colours_for('parents', take_first=False, verbose=True)
        print(tp)
        self.assertEqual(len(tp), 4)

    def test_vfbterm_site(self):
        term = self.vfb.term('fafb_catmaid_api')
        print("got term ", term)
        term.debug = True
        term._return_type='id'
        print(term.instances)
        self.assertGreater(len(term._instances_ids), 9)
        print(dir(term))
        print(term.get_summary(return_dataframe=False))
        self.assertTrue(term.has_image)

    def test_vfbterm_xref(self):
        term = self.vfb.term('VFB_jrcv1ngs')
        print("got term ", term)
        print(term.xrefs)
        self.assertTrue(term.xref_id)
        print(dir(term))
        print(term.xref_id)
        self.assertEqual(self.vfb.xref_2_vfb_id(term.xref_id, return_just_ids=True)[0], term.id)

if __name__ == "__main__":
    unittest.main()


#!/usr/bin/env python3
import pickle
import requests
import json
import re
import time
from datetime import timedelta
import math
import argparse
from ..default_servers import get_default_servers
import os


def cli_credentials():
    """Parses command line credentials for Neo4J rest connection;
    Optionally specify additional args as a list of dicts with
    args required by argparse.add_argument().  Order in list
    specified arg order"""
    parser = argparse.ArgumentParser()
    parser.add_argument("endpoint",
                    help="Endpoint for connection to neo4J prod")
    parser.add_argument("usr",
                    help="username")
    parser.add_argument("pwd",
                    help="password")
#    if additional_args:
#        for a in additional_args:
#            parser.add_argument(**a)  # how to get this to work with non kewyord args
    return parser.parse_args()

def cli_neofj_connect():
    args = cli_credentials()
    return Neo4jConnect(endpoint=args.endpoint,
                        usr=args.usr,
                        pwd=args.pwd)


def chunks(l, n):
    """Yield successive n-sized chunks from l.

    :param l: a list
    :param: n, chunk size"""
    for i in range(0, len(l), n):
        yield l[i:i+n]


# @dataclass
# class Filter:
#     jpath: str
#     label: str
#     value_restriction: None
#
# @dataclass
# class CompoundFilter:
#     primary_filter: Filter
# #    secondary_filters: Array(Filter)
#
# #    image_folder_plus_meta = Filter(jpath="$.channel_image.[*].image.image_folder,template_anatomy")


# def batch_query_dict_opt(func):
#      # Assumes first arg is to be batched and that return value is dict
#      def wrapper_batch(*args, **kwargs):
#          if not (args[1] is None):
#              cs = chunks(args[1], 1000)
#          else:
#              return func(*args, **kwargs)
#          out = dict()
#          for c in cs:
#              arglist = list(args)
#              arglist[1] = c
#              subargs = tuple(arglist)
#              out.update(func(*subargs, **kwargs))
#      return wrapper_batch


class Neo4jConnect:
    """Thin layer over REST API to hold connection details,
    handle multi-statement POST queries, return results and report errors.

    :param endpoint: a neo4j REST endpoint
    :param usr: username (content ignored if credentials not rqd)
    :param pwd: password (content ignored if credentials not rqd)"""
    # Return results might be better handled in the case of multiple statements - especially when chunked.
    # Not connection with original query is kept.

    def __init__(self, endpoint = get_default_servers()['neo_endpoint'],
                 usr=get_default_servers()['neo_credentials'][0],
                 pwd=get_default_servers()['neo_credentials'][1]):
        self.base_uri = endpoint
        self.usr = usr
        self.pwd = pwd
        self.commit = "/db/neo4j/tx/commit"
        self.headers = {'Content-type': 'application/json'}
        if not self.test_connection():
            print("Falling back to Neo4j v3 connection")
            self.commit = "/db/data/transaction/commit"
            self.headers = {}
            if not self.test_connection():
                raise Exception("Failed to connect to Neo4j.")

    def commit_list(self, statements, return_graphs=False):
        """Commit a list of statements to neo4J DB via REST API.
        Errors prompt warnings (STDERR), not exceptions, and cause return = FALSE.

        :param: statements: A list of cypher statements.
        :param: return_graphs: Optional. Boolean. Returns graphs under 'graph' key if True. Default: False
        :Return: List of results or False if any errors are encountered.
        """

        cstatements = []
        if return_graphs:
            for s in statements:
                cstatements.append({'statement': s, "resultDataContents" : [ "row", "graph" ]})
        else:
            for s in statements:
                cstatements.append({'statement': s}) # rows an columns are returned by default.
        payload = {'statements': cstatements}
        try:
            response = requests.post(url = "%s%s"
                                 % (self.base_uri, self.commit), auth = (self.usr, self.pwd) ,
                                  data = json.dumps(payload), headers = self.headers)
        except requests.exceptions.RequestException as e:
            print("\033[31mConnection Error:\033[0m %s" % e)
            print("Retrying in 10 seconds...")
            time.sleep(10)
            return self.commit_list(statements)
        if self.rest_return_check(response):
            return response.json()['results']
        else:
            return False

    def commit_list_in_chunks(self, statements, verbose=False, chunk_length=1000):

        """Commit multiple (chunked) commit of statements to neo4J DB via REST API.
         Errors prompt warnings (STDOUT), not exceptions, and cause return = FALSE.

         :param: statements: A list of cypher statements.
         :param: verbose: Boolean. Optionally print periodic reports of progress to STDOUT
         :param: chunk_length. Int. Optional. Set chunk size.  Default = 1000.
         :Return: List of results or False if any errors are encountered. Chunking is not reflected in results.
         """

        chunked_statements = chunks(l = statements, n=chunk_length)
        chunk_results = []
        i = 1
        c_no = math.ceil(len(statements)/chunk_length)
        for c in chunked_statements:
            if verbose:
                start_time = time.time()
                print("Processing chunk of %d of %d starting with: %s" % (i,
                                                                          c_no,
                                                                          c[0].encode('utf8')))
            r = self.commit_list(c)
            if verbose:
                t = time.time() - start_time
                print("Processing took %d seconds for %s statements" % (t, len(c)))
                print("Estimated time to completion: %s." % str(timedelta(seconds=(t*(c_no - i)))))
            if type(r) == list:
                chunk_results.extend(r)
            else:
                chunk_results.append(r)
            i += 1
        return chunk_results

    def commit_csv(self, url, statement, chunk_size=1000, sep=","):
        # May need some configuration to work with file://...
        cypher = "USING PERIODIC COMMIT %d " \
                 "LOAD CSV WITH HEADERS FROM '%s' AS line FIELDTERMINATOR '%s' " \
                 "%s" % (chunk_size, url, sep, statement)
        self.commit_list([cypher])

    def rest_return_check(self, response):
        """Checks status response to post. Prints warnings if not OK.
        If OK, checks for errors in response. Prints any present as warnings.
        Returns True STATUS OK and no errors, otherwise returns False.
        """
        if not (response.status_code == 200):
            print("\033[31mConnection Error:\033[0m %s (%s)" % (response.status_code, response.reason))
            return False
        else:
            j = response.json()
            if j['errors']:
                for e in j['errors']:
                    print("\033[31mQuery Error:\033[0m " + str(e))
                return False
            else:
                return True

    def test_connection(self):
        """Test neo4j endpoint connection"""
        statements = ["MATCH (n) RETURN n LIMIT 1"]
        if self.commit_list(statements):
            return True
        else:
            return False

    def list_all_node_props(self):
        r = self.commit_list(['MATCH (n) with keys(n) AS kl UNWIND kl as k RETURN DISTINCT k'])
        d = dict_cursor(r)
        return [x['k'] for x in d]

    def list_all_edge_props(self):
        r = self.commit_list(['MATCH ()-[r]-() with keys(r) AS kl UNWIND kl as k RETURN DISTINCT k'])
        d = dict_cursor(r)
        return [x['k'] for x in d]

    def get_lookup(self, cache=None, limit_type_by_prefix=('FBbt', 'VFBexp', 'VFBext'), include_individuals=True,
                   limit_properties_by_prefix=('RO', 'BFO', 'VFBext'), curies=False, include_synonyms=True, verbose=False):
        """Generate a name:ID lookup from a VFB neo4j DB, optionally restricted by a list of prefixes.

        :param cache: If a valid cache path is provided, uses a cached lookup. Otherwise, generates a new lookup.
        :param limit_type_by_prefix: Optional list of id prefixes for limiting lookup of classes & individuals.
        :param include_individuals: If `True`, individuals are included in the lookup.
        :param limit_properties_by_prefix: Optional list of id prefixes for limiting lookup of properties.
        :param curies: If `True`, returns CURIEs instead of IDs.
        :param include_synonyms: If `True`, includes synonyms in the lookup.
        :param verbose: If `True`, provides verbose output.
        :return: A dictionary with names (or synonyms) as keys and their corresponding IDs as values.
        """
        try:
            three_months_in_seconds = 3 * 30 * 24 * 60 * 60
            if cache and os.path.exists(cache) and os.path.getctime(cache) > time.time() - three_months_in_seconds:
                if verbose:
                    print("Loading cache lookup from disk...")
                with open(cache, 'rb') as f:
                    return pickle.load(f)
        except Exception as e:
            print(f"Failed to load cache lookup from disk: {e}")

        print("Caching all terms for faster lookup...")

        out = []

        # Step 1: Load Classes
        self.load_classes(out, limit_type_by_prefix, include_synonyms, verbose)

        # Step 2: Load Individuals
        if include_individuals:
            self.load_individuals(out, include_synonyms, verbose)

        # Step 3: Load ObjectProperties
        self.load_object_properties(out, limit_properties_by_prefix, verbose)

        # Remove duplicates and convert to final output format
        lookup = self.process_results(out, curies)

        # Save the lookup to a cache file
        self.save_to_cache(lookup, cache)

        return lookup

    def load_classes(self, out, limit_type_by_prefix, include_synonyms, verbose):
        """Load Class terms into the output list.

        :param out: The list to store fetched terms.
        :param limit_type_by_prefix: Optional list of id prefixes for limiting lookup of classes.
        :param include_synonyms: If `True`, includes synonyms in the lookup.
        :param verbose: If `True`, provides verbose output.
        """
        where_clause = self.construct_where_clause(limit_type_by_prefix, "AND NOT a:Deprecated")
        self.execute_and_process_query('Class', where_clause, 'label', out, verbose)
        self.execute_and_process_query('Class', where_clause, 'symbol[0]', out, verbose)

        if include_synonyms:
            self.execute_and_process_query_with_synonyms('Class', where_clause, out, verbose)

    def load_individuals(self, out, include_synonyms, verbose):
        """Load Individual terms into the output list.

        :param out: The list to store fetched terms.
        :param include_synonyms: If `True`, includes synonyms in the lookup.
        :param verbose: If `True`, provides verbose output.
        """
        where_clause = "AND NOT a:Deprecated AND NOT a.short_form STARTS WITH 'VFBc_' AND NOT a:Person"
        self.execute_and_process_query('Individual', where_clause, 'label', out, verbose)
        self.execute_and_process_query('Individual', where_clause, 'symbol[0]', out, verbose)

        if include_synonyms:
            self.execute_and_process_query_with_synonyms('Individual', where_clause, out, verbose)

    def load_object_properties(self, out, limit_properties_by_prefix, verbose):
        """Load ObjectProperties into the output list.

        :param out: The list to store fetched terms.
        :param limit_properties_by_prefix: Optional list of id prefixes for limiting lookup of properties.
        :param verbose: If `True`, provides verbose output.
        """
        where_clause = self.construct_where_clause(limit_properties_by_prefix)
        self.execute_and_process_query('ObjectProperty', where_clause, 'label', out, verbose)

        # Handle alternative terms for ObjectProperties
        if limit_properties_by_prefix:
            match_string = "' OR a.short_form STARTS WITH '".join(limit_properties_by_prefix)
            where_clause = f"WHERE EXISTS(a.alternative_term) AND size(a.alternative_term) > 0 AND a.short_form STARTS WITH '{match_string}'"
        else:
            where_clause = "WHERE EXISTS(a.alternative_term) AND size(a.alternative_term) > 0"
        
        query = f"MATCH (a:ObjectProperty) {where_clause} UNWIND a.alternative_term as label RETURN a.short_form as id, label as name"
        q = self.commit_list([query])
        self.process_and_add_results(q, out, verbose)

    def construct_where_clause(self, prefixes, base_clause=""):
        """Construct the WHERE clause for the query based on term type and prefix limitations.

        :param prefixes: List of id prefixes for limiting lookup.
        :param base_clause: Additional base condition for the WHERE clause.
        :return: A string representing the WHERE clause for the query.
        """
        if prefixes:
            regex_string = '.+|'.join(prefixes) + '.+'
            where = f"{base_clause} AND a.short_form =~ '{regex_string}'"
        else:
            where = base_clause
        return where

    def execute_and_process_query(self, term_type, where, property_name, out, verbose):
        """Execute a Cypher query and process the results.

        :param term_type: The type of terms to fetch (e.g., 'Class', 'Individual', 'ObjectProperty').
        :param where: The WHERE clause of the query.
        :param property_name: The property to fetch (e.g., 'label', 'symbol[0]').
        :param out: The list to store fetched terms.
        :param verbose: If `True`, provides verbose output.
        """
        query = f"MATCH (a:{term_type}) WHERE EXISTS(a.short_form) {where} AND EXISTS(a.{property_name.split('[')[0]}) RETURN a.short_form as id, a.{property_name} as name ORDER BY id DESC"
        q = self.commit_list([query])
        self.process_and_add_results(q, out, verbose)

    def execute_and_process_query_with_synonyms(self, term_type, where, out, verbose):
        """Execute a Cypher query to fetch terms and their synonyms.

        :param term_type: The type of terms to fetch (e.g., 'Class', 'Individual').
        :param where: The WHERE clause of the query.
        :param out: The list to store fetched terms.
        :param verbose: If `True`, provides verbose output.
        """
        query = f"""
            MATCH (a:{term_type}) WHERE EXISTS(a.short_form) {where} AND (EXISTS(a.synonyms) OR (a)-[:has_reference {{typ:'syn'}}]->(:pub:Individual)) 
            UNWIND a.synonyms AS synonym2 
            RETURN DISTINCT a.short_form AS id, synonym2 AS name 
            UNION ALL 
            MATCH (a:{term_type})-[r:has_reference {{typ:'syn'}}]->(:pub:Individual) 
            UNWIND r.value AS synonym1 
            WITH a.short_form AS id, synonym1 AS synonym 
            RETURN DISTINCT id, synonym AS name
        """
        q = self.commit_list([query])
        self.process_and_add_results(q, out, verbose)

    def process_and_add_results(self, query_result, out, verbose):
        """Process the results of a query and add to the output list, checking for duplicates.

        :param query_result: Result set from the query.
        :param out: The list to store fetched terms.
        :param verbose: If `True`, provides verbose output.
        """
        # Create a dictionary to store existing names and their corresponding IDs
        name_to_id = {item['name']: item['id'] for item in out}

        # Iterate through the query results
        for result in dict_cursor(query_result):
            name = result['name']
            id = result['id']

            # Check if the name is already in the list and if the current term should take priority
            if name in name_to_id:
                existing_id = name_to_id[name]
                if verbose:
                    print(f"Skipping {name} with ID {id} as it is already represented by existing term {existing_id}")
                continue
            else:
                out.append(result)
                name_to_id[name] = id

    def process_results(self, out, curies):
        """Remove duplicates and prepare the final lookup dictionary.

        :param out: List of results to process.
        :param curies: If `True`, convert IDs to CURIE format.
        :return: Final lookup dictionary.
        """
        seen = set()
        unique_out = []
        for item in out:
            pair = (item['name'], item['id'])
            if pair not in seen:
                seen.add(pair)
                unique_out.append(item)

        if curies:
            lookup = {x['name']: x['id'].replace('_', ':') for x in unique_out}
            lookup.update({x['id'].replace(':', '_'): x['id'] for x in unique_out})
        else:
            lookup = {x['name']: x['id'] for x in unique_out}

        return lookup

    def save_to_cache(self, lookup, cache):
        """Save the lookup to a cache file if caching is enabled.

        :param lookup: The lookup dictionary to save.
        :param cache: The cache file path.
        """
        if cache:
            try:
                with open(cache, 'wb') as f:
                    pickle.dump(lookup, f)
            except Exception as e:
                print(f"Failed to save cache lookup to disk: {e}")

def dict_cursor(results):
    """Takes JSON results from a neo4J query and turns them into a list of dicts.

    :param results: neo4j query results
    :return: list of dicts
    """
    dc = []
    for n in results:
        # Add conditional to skip any failures
        if n:
            for d in n['data']:
                dc.append(dict(zip(n['columns'], d['row'])))
    return dc


def escape_string(strng):
    """Simple escaping for strings used in neo queries."""
    if type(strng) == str:
        strng = re.sub(r'\\', r'\\\\', strng)
        strng = re.sub("'", "\\'", strng)
        strng = re.sub('"', '\\"', strng)
    return strng





from setuptools import setup, find_packages
from os import path
import glob

here = path.abspath(path.dirname(__file__))

from get_version import get_version
__version__ = get_version(__file__)
del get_version

# Get the long description from the README file
with open(path.join(here, 'README.md')) as f:
    long_description = f.read()

setup(name='vfb_connect',  # Required
      version=__version__,  # Required
      packages=find_packages(where='src'),
      package_dir={'': 'src'},
      py_modules=[path.splitext(path.basename(path))[0] for path in glob.glob('src/*.py')],
      include_package_data=True,
      description='Wrapper for querying VirtualFlyBrain servers.',  # Optional)
      long_description=long_description,
      long_description_content_type='text/markdown',
      author='David Osumi-Sutherland',  # Optional
      url='https://github.com/VirtualFlyBrain/VFB_connect',
      # This should be a valid email address corresponding to the author listed
      # above.
      author_email='dosumis@gmail.com',  # Optional
      install_requires=['requests', 'pandas', 'jsonpath_rw', 'pysolr', 'navis', 'flybrains', 'numpy>=1.26.4, <2.0.0', 'neuprint-python', 'python-catmaid', 'matplotlib>3.9', 'seaborn>0.13', 'colormath', 'tqdm', 'vispy', 'ipywidgets'],
      data_files=[('json_schema', ['src/vfb_connect/resources/VFB_TermInfo_queries.json','src/vfb_connect/resources/VFB_results_single_input.json','src/vfb_connect/resources/VFB_results_multi_input.json'])],
      classifiers=[  # Optional
          # How mature is this project? Common values are
          #   3 - Alpha
          #   4 - Beta
          #   5 - Production/Stable
          'Development Status :: 4 - Beta',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Programming Language :: Python :: 3',
       ],
      project_urls={  # Optional
          'Bug Reports': 'https://github.com/VirtualFlyBrain/VFB_connect/issues',
          'Source': 'https://github.com/VirtualFlyBrain/VFB_connect',
          'Documentation': 'https://vfb-connect.readthedocs.io/en/stable/'
       },
)
